
/*
Copyright (c) 2019 The Khronos Group Inc.
Use of this source code is governed by an MIT-style license that can be
found in the LICENSE.txt file.
*/


#ifdef GL_ES
precision mediump float;
#endif
varying struct s {
    int i;
} s1;  // structures cannot be declared with varying qualifier

void main()
{
}
